package org.wakeland.pedro;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.ConnectException;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.UnknownHostException;
import java.text.SimpleDateFormat;
import java.util.Calendar;

import javax.imageio.stream.FileImageOutputStream;
import javax.microedition.io.Connector;
import javax.microedition.io.StreamConnectionNotifier;
import javax.obex.HeaderSet;
import javax.obex.Operation;
import javax.obex.ResponseCodes;
import javax.obex.ServerRequestHandler;
import javax.obex.SessionNotifier;

/**
 * 
 * Class acting as a Bluetooth server, waiting for connection over OBEX protocol 
 * from MobileImagineConnectorClient running on mobile phone.
 * 
 * <p>Original sample code of OBEX server downloaded from
 * <a href="http://wiki.forum.nokia.com/index.php/Using_Bluetooth_obex_server_in_Java_ME">Nokia wiki forum</a>.</p>
 * 
 * @author Peter (Pedro) Nosal, pedron4@gmail.com
 * @version 1.0
 *
 */
public class MobileImagineConnectorServer implements Runnable {

    /**
     * Server state flag initialized to 'not running' e.g. false.
     */
    boolean serverStarted = false;
    /**
     * String buffer for messages.
     */
    StringBuffer msgSendBuffer;
    /**
     * Thread that listens for incoming connections.
     */
    private Thread threadMain;
    /**
     * Stores a reference to incoming connection notifier.
     */
    private SessionNotifier sessionNotifier;
    /**
     * Used for handling client requests (GET and PUT).
     * @see RequestHandler
     * @see ServerRequestHandler
     */
    private RequestHandler handler;
	/**
     * Constant for message type - text.
     */
    public final static int MESSAGE_TYPE_TEXT = 1;
    /**
     * Constant for message type - pressed button.
     */
    public final static int MESSAGE_TYPE_BUTTON = 2;
    /**
     * Constant for message type - image.
     */
    public final static int MESSAGE_TYPE_IMAGE = 3;
    /**
     * Date and time format used in saving image.
     */
	private final static String DATE_FORMAT_NOW = "yyyy-MM-dd_HH-mm";

    //
    // 
    // SOCKET CONNECTION variables
    // 
    //
    /**
     * Instance of Socket - for creating connection with Imagine Logo through serial port. 
     */
	private Socket socket;
	/**
	 * Port on which this application connects with Imagine Logo.
	 */
	private static final int PORT = 5151;
	/**
	 * OuputStream for receiving data from Imagine Logo.
	 */
	private OutputStream wr;
	/**
	 * InputStream for sending data to Imagine Logo.
	 */
	private InputStream rd;
	/**
	 * Nickname of this application given by Imagine Logo.
	 */
	private String nickname;



	/**
	 * Constructs new MobileImagineConnectorServer class.
	 * Initializes instance of RequestHandler that listens for OBEX requests.
	 */
	public MobileImagineConnectorServer(){
		handler = new RequestHandler();
		msgSendBuffer = new StringBuffer();

		startObexServer();
	}

	/**
	 * Returns current minute, hour, day, month and year if format set in constant DATE_FORMAT_NOW.
	 * @return	current minute, hour, day, month and year if format set in constant DATE_FORMAT_NOW.
	 */
	public String getTime() {
		Calendar cal = Calendar.getInstance();
		SimpleDateFormat sdf = new SimpleDateFormat(DATE_FORMAT_NOW);
		return sdf.format(cal.getTime());
	}

    /**
     * Method creates a connection notifier and initializes a connection
     * waiting thread.
     * @return 		0 if server startup was successful. 1 if connection notifier
     * instantiation caused an exception. 2 if thread creating caused an
     * exception.
     * @see Connector#open(java.lang.String)
     * @see Thread#start() 
     * @see MobileImagineConnectorServer#run() 
     */
    private int startObexServer() {
    	String uuid = "49C41BB556414702B914EFEE734AF38B";
        String strServerConnection = "btgoep://localhost:" + uuid + ";name=Mobile Imagine Connector;master=true";
        try {
            System.out.println("Starting server...");
            sessionNotifier = (SessionNotifier) Connector.open(strServerConnection);
            threadMain = new Thread(this);
            threadMain.start();
        } catch ( IllegalThreadStateException e ) {
            System.out.println("Error starting a server thread!");
            return 2;
 
        } catch ( Exception e ) {
            System.out.println("Error starting a server connection!");
            return 1;
        }
        return 0;
    }

    /**
     * From Runnable:
     * The method is being run when a connection listening thread was created
     * and started(in startObexServer). It runs acceptAndOpen() method.
     * @see StreamConnectionNotifier
     */
    public void run() {
        try {
            System.out.println("Waiting for connections...");
            sessionNotifier.acceptAndOpen(handler);
            System.out.println("Incomming connection...");
            //waiting for messages
        } catch ( Exception e ) {
        	System.out.println("Connection waiting thread interrupted!");
        	System.out.println(e.toString());
        }
        // this part was added to keep the server running
        boolean x = true;
        do {
        	try {
        		Thread.sleep(1000);
        	} catch (InterruptedException e) {}
        } while (x);
        System.out.println(Boolean.toString(handler.equals("x")));
        System.out.println(Boolean.toString(sessionNotifier.equals("x")));
    }

    /**
     * Gets bytes from msgSendBuffer. Method cleans buffer data after execution
     * if the buffer was not empty.
     * @return 	Array of bytes retrieved from msgSendBuffer.
     */
    private byte[] popMsgFromSendBuf() {
        // getting data from message buffer
        byte[] arrMsgBytes = null;
        synchronized ( msgSendBuffer ) {
            arrMsgBytes = msgSendBuffer.toString().getBytes();
            if (arrMsgBytes.length > 0) {
                msgSendBuffer.delete( 0, msgSendBuffer.length() );
            }
        }
        return arrMsgBytes;
    }

    /**
     * Puts message to msgSendBuffer.
     * @param message 	The message to be sent to MobileImagineConnectorServer.
     * @see MobileImagineConnectorServer#msgSendBuffer
     */
    private void putMsgToSendBuffer(String message) {
        msgSendBuffer.append(message);
    }

    
    
    
    /**
     * Defines an event listener that will respond to OBEX requests made to 
     * the server.
     * @see ServerRequestHandler
     */
    private class RequestHandler extends ServerRequestHandler {

    	/**
         * Default constructor.
         */
        public RequestHandler() {
           // no implementation required. 
        }

        /**
         * From ServerRequestHandler:
         * Sends data to the client that initialized the GET request.
         * 
         * Checks whether msgSendBuffer is empty right now.
         * If the buffer is empty than OBEX_HTTP_NO_CONTENT will be returned to
         * the client.
         * If the buffer contains some data than this method performs a following
         * procedure:
         * <ul>
         * <li>current msgSendBuffer content is converted to byte[],</li>
         * <li>gets current operation header(HeaderSet),</li>
         * <li>sets header TYPE to BtObexMessage,</li>
         * <li>sets header LENGTH equal to length of the byte array containing msgSendBuffer content,</li>
         * <li>opens openDataOutputStream for the current operation and writes the byte array there,</li>
         * <li>finally OBEX_HTTP_ACCEPTED is being returned.</li>
         * </ul>
         * 
         * @param op 	Contains the headers sent by the client.
         * @return OBEX_HTTP_OK 	If the request handling procedure executed.
         * OBEX_HTTP_NO_CONTENT 	If no message found in the msgSendBuffer.
         */
        public int onGet(Operation op) {
        	OutputStream out = null;
            HeaderSet header = null;
            byte[] arrMsgBytes = null;
 
            arrMsgBytes = popMsgFromSendBuf();
            if (arrMsgBytes.length == 0) {
                return ResponseCodes.OBEX_HTTP_NO_CONTENT;
            }
            //creating a header to send back to the client
            header = createHeaderSet();
            //adding our headers
            //TYPE - represents obex message
            header.setHeader( HeaderSet.TYPE, "BtObexMessagea" );
            //LENGTH - holds message length
            header.setHeader(HeaderSet.LENGTH, new Long(arrMsgBytes.length));
            try {
                op.sendHeaders(header);
            } catch (Exception ex) {
                System.out.println("Error during sending a message to client!");
            }
            try {
                // Open the output stream to put the object to it
                out = op.openDataOutputStream();
 
                // Sending the object to client
                out.write(arrMsgBytes);
                out.flush();

                // printing the text to be sent to client to the standard output
                String str = "";
                for (int i = 0; i < arrMsgBytes.length; i++)
                	str += (char)arrMsgBytes[i];
                System.out.println("Sending message to client: " + str);

            } catch (Exception e) {
                System.out.println("Error during sending a message to client!");
            } finally {
                // End the transaction
                //closing data stream
                if (out != null) {
                    try {
                        out.close();
                    } catch (Exception e) {
                        System.out.println("Error during sending a message to client!");
                    }
                }
            }
            return ResponseCodes.OBEX_HTTP_OK;
        }

        /**
         * From ServerRequestHandler:
         * Receives data from the client that initialized the PUT request.
         * 
         * Checks whether header(HeaderSet) is present in the operation.
         * If it doesn't than OBEX_HTTP_BAD_REQUEST will be send back to
         * the client that initiated a PUT request.
         * If a header is detected than function performs the following procedure:
         * <ul>
         * <li>openInputStream is opened for current operation,</li>
         * <li>function reads the first byte to recognize the message type,</li>
         * <li>according to the message type the function decides what to do next,</li>
         * <li>finally OBEX_HTTP_OK is being returned.</li>
         * </ul>
         * 
		 * WARNING!!! If the length of the message received is less than 1024 bytes,
		 * the onDelete() method is being called, not the onPut() method. To make sure
		 * the onPut() method is being called, message with length less than 1024 bytes
		 * is filled by 0's (zeros) to have 1024 bytes.
         * 
         * @param op 	Contains the headers sent by the client.
         * @return OBEX_HTTP_BAD_REQUEST 	If no header(HeaderSet) detected and
         * 		   OBEX_HTTP_OK 			if the request handling procedure executed.
         */
        public int onPut(Operation op) {
            HeaderSet header = checkHeader(op);
            if (header == null)
                return ResponseCodes.OBEX_HTTP_BAD_REQUEST;

            try {
                // opening stream to read the incoming message
                InputStream is = op.openInputStream();

                // reading data
                StringBuffer buf = new StringBuffer();
                int data;
                // reading message type
                int messageType = 0;
                messageType = is.read() - 48;

                char c = ' ';
                switch (messageType){
                case MESSAGE_TYPE_BUTTON:
                case MESSAGE_TYPE_TEXT:
                	// reading message length
                	int messageLength = 0;
                	while ((data = is.read()) != '*'){
                		messageLength *= 10;
                		messageLength += (data - 48);
                	}
                	// reading as many chars as messageLength
                	int counter = 0;
                	while ((data = is.read()) != -1 && counter < messageLength) {
                		buf.append((char)data);
                		if (c == ' ')
                			c = (char)data;
                		counter++;
                	}

                	System.out.println("Received message from client: " + buf.toString());
                	// if there is active connection with Imagine,
                	// send received message to Imagine
                	if (isConnectedToImagine())
                		sendMsgText(buf.toString());

                	break;
                case MESSAGE_TYPE_IMAGE:
                	System.out.println("Receiving image from client...");
                	// reading all data
                    while ((data = is.read()) != -1){
                        buf.append((char)data);
                    }

                    File outputfile = new File("image_" + getTime() + ".jpeg");
                    FileImageOutputStream fios = new FileImageOutputStream(outputfile);
                    fios.writeBytes(buf.toString());
                    fios.close();

                    System.out.println("Image received from client.");
                	// if there is active connection with Imagine,
                    // send path to image to Imagine
                	if (isConnectedToImagine())
                		sendMsgText(outputfile.getAbsolutePath());

                	break;
                case 0:
                	break;
                default:
                	System.out.println("Unknown message type: " + messageType);
                	break;
                }

                // closing operation
                op.close();

                return ResponseCodes.OBEX_HTTP_OK;
            } catch (IOException e) {
                e.printStackTrace();
                return ResponseCodes.OBEX_HTTP_UNAVAILABLE;
            }
        }

        /**
         * From ServerRequestHandler:
         * Does nothing, if this method is called it is the fault of the implementation.
         * 
         * <p>If this method is called, it is probably a fault of
         * the OBEX protocol implementation.</p>
         * 
         * @param request	
         * @param reply		
         */
        public int onDelete(HeaderSet request, HeaderSet reply) {
        	System.out.println("### onDelete() method called - FAULT of the OBEX implementation.");
            return ResponseCodes.OBEX_HTTP_OK;
        }

        /**
         * Gets header(HeaderSet) from the operation and returns a pointer to
         * it. If there was an error function returns null.
         * @param op 		Contains the headers sent by the client.
         * @return 			Null if header getting procedure threw and exception,
         * 					a pointer to HeaderSet object will be returned in case of success.
         */
        protected HeaderSet checkHeader(Operation op) {
            HeaderSet header = null;
            try {
                header = op.getReceivedHeaders();
            } catch (Exception e) {
                System.out.println("Error getting request header!");
                return null;
            }
            return header;
        }

        /**
         * From ServerRequestHandler:
         * Called when a DISCONNECT request is received.
         * @param request 	Contains the headers sent by the client; request will never be null.
         * @param reply 	The headers that should be sent in the reply; reply will never be null.
         */
        public void onDisconnect(HeaderSet request, HeaderSet reply) {
            System.out.println("Disconnecting...");
        }

        /**
         * From ServerRequestHandler:
         * Called when a CONNECT request is received.
         * @param request 	Contains the headers sent by the client; request will never be null.
         * @param reply 	The headers that should be sent in the reply; reply will never be null.
         * @return		OBEX_HTTP_OK when the connection is established.
         */
        public int onConnect(HeaderSet request, HeaderSet reply) {
            System.out.println("Client is connecting...");
            String message = "aGreetings from PC";
    	    putMsgToSendBuffer(message);
    	    // starting thread that waits for Imagine to connect
    	    startSocketConnection();
            return ResponseCodes.OBEX_HTTP_OK;
        }
    }

    /**
     * Creates new instance of SocketConn - class for making
     * serial port connection with Imagine Logo.
     */
    private void startSocketConnection(){
    	new SocketConn();
    }

    /**
     * Class for making serial port connection on port 5151 with Imagine Logo.
     * 
     * @author Peter (Pedro) Nosal, pedron4@gmail.com
     * @version 1.0
     *
     */
    public class SocketConn extends Thread {

    	/**
    	 * Constructor creates and starts a new thread.
    	 */
    	public SocketConn(){
    		Thread th = new Thread(this);
    		th.start();
    	}

    	/**
    	 * From Thread:
    	 * Calls createConnection() to open the serial port and checks whether there is still
    	 * an active connection with Imagine.
    	 */
    	public void run(){
    		createConnection();

    		while (isConnectedToImagine()){
    			receiveMessage();
    		}
    		System.out.println("Error: connection with Imagine lost!");
    		try {
    			socket.close();
    		} catch (Exception e) {
    			e.printStackTrace();
    		}
    	}
    }

    /**
     * Returns whether there is an active connection with Imagine.
     * @return		True if connection is active, false otherwise.
     */
    private boolean isConnectedToImagine(){
    	return socket.isConnected();
    }

    /**
     * Initializes Socket class on port in constant PORT and either waits for the Imagine to connect,
     * or connects to Imagine, if the Imagine is running and waiting for this application to connect.
     */
    private void createConnection()	{
		try {
		    try {
		    	// first try to connect to other party, if it is already listening
			    socket = new Socket("localhost", PORT);
			    System.out.println("Socket for sending to Imagine created on port " + PORT + ".");
		    } catch (ConnectException e) {
		    	// otherwise create a listening socket and wait for the other party to connect
			    System.out.println("Waiting for Imagine...");
		    	ServerSocket srv = new ServerSocket(PORT);
			    socket = srv.accept();
			    System.out.println("Socket for receiving from Imagine created on port " + PORT + ".");
			}
		    wr = socket.getOutputStream();
            rd = socket.getInputStream();
		} catch (UnknownHostException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

    /**
     * Puts the string in the right format and calls sendMsg() to send this message to Imagine.
     * @param message	The message to sent to Imagine.
     */
	private void sendMsgText(String message){
		sendMsg("w1!" + nickname.length() + "!" + nickname + message);
	}

	/**
	 * Sends message that has to be already formated to the Imagine.
	 * @param message	The message to sent to Imagine.
	 */
	public void sendMsg(String message) {
		if (isConnectedToImagine()){
			message = message.length() + "!" + message;
			System.out.println("Sending message to Imagine: " + message);
			byte bts[] = message.getBytes();
			try {
				wr.write(new byte[] {7});
				wr.write(bts);
				wr.flush();
			} catch (IOException e) {
				e.printStackTrace();
			}
		} else {
			System.out.println("Error: unable to send message to Imagine, connection with Imagine lost!");
		}		
	}

	/**
	 * Receives message from Imagine.
	 * @return	False if IOException occurred during reading, true otherwise.
	 */
	public boolean receiveMessage() {
		try {
			// skip the first character
			rd.read();

			StringBuffer buf = new StringBuffer();
			int data;
			int counter = 0;
			while ((data = rd.read()) != '!'){
				counter *= 10;
				counter += data - 48;
			}
			while (counter > 0){
				data = rd.read();
				buf.append((char)data);
				counter--;
			}
			if (buf.length() > 0)
				parseMessage(buf.toString());
		} catch (IOException e) {
			System.out.println(e.toString());
			return false;
		}
		return true;
	}

    /**
     * Parses the message that came from Imagine.
     * Imagine protocol is defined in the net_protocol.txt.
     * 
     * Looks up for the first letter and then decides what to do.
     * Set of commands (depending on what the first letter is):
     * <ul>
     * <li>a: login request with a specified nickname. We respond to this request.</li>
     * <li>W: regular message. This way Imagine sends commands to the mobile phone through this server.</li>
     * </ul>
     * @param message	The message that came from Imagine.
     */
    private void parseMessage(String message) {
    	switch (message.charAt(0)){
    	case 'a':
    		// we determine the nickname
    		nickname = message.substring(message.indexOf('!')+1);
    		sendMsg("b0!"+nickname);
    		System.out.println("Connection with Imagine successfully established!");
    		break;
    	case 'W':
    		// decoding the message
    		int indexStart = 4 + nickname.length() + String.valueOf(nickname.length()).length();
    		int tmp = 0;
    		for (int i = indexStart; i < message.indexOf('!', 10); i++){
    			tmp *= 10;
    			tmp += message.charAt(i) - 48;
    		}
    		String msg = message.substring(message.indexOf('!', indexStart) + 1 + tmp);
    		System.out.println("Received message from Imagine: " + msg);
    		putMsgToSendBuffer(msg);
    		break;
    	default:
    		System.out.println("default: " + message);
    	}
	}


    /**
     * Main method for creating new MobileImagineConnectorServer object.
     * @param args	Empty.
     */
    public static void main(String[] args){
    	new MobileImagineConnectorServer();
    }
}